/*
 * RC4
 * 
 * (!c) 2001 by Ighighi
 * Venezuela
 * 
 * Based on code posted to sci.crypt by an anonymous cypherpunk
 * 
 * RC4 is a variable key-size (up to 2048 bits) stream cipher
 * developed in 1987 by Ron Rivest of RSA Data Security, Inc.
 */

#include "rc4.h"

void RC4_Init(RC4 *rc4, const unsigned char *key, int len)
{
    int x, y;
    int i;
    unsigned char *state;
    unsigned char tmp;

    if (! (rc4 && key))
	return;
    
    state = &rc4->state[0];         
    for (i = 0; i < 256; i++)
        state[i] = i;

    rc4->x = rc4->y = 0;
    x = y = 0;    
    
    for (i = 0; i < 256; i++) {
	tmp = state[i];
	y = (key[x] + tmp + y) & 0xff;	/* y = (key[x] + tmp + y) % 256; */
	if (++x == len) x = 0;		/* x = (x + 1) % len; */
	state[i] = state[y];
	state[y] = tmp;
    }
    
    return;
}

void RC4_Crypt(RC4 *rc4, unsigned char *inbuf, unsigned char *outbuf, int len)
{ 
    int x, y;
    int i;
    int xorIndex;
    unsigned char *state;
    unsigned char tmp;

    if (! (rc4 && inbuf && outbuf))
	return;
    
    x = rc4->x;
    y = rc4->y;     
    
    state = &rc4->state[0];
    for (i = 0; i < len; i++) {
	x = (x + 1) & 0xff;
	tmp = state[x];
	y = (tmp + y) & 0xff;
	state[x] = state[y];
	state[y] = tmp;
	xorIndex = (state[x] + state[y]) & 0xff;
	outbuf[i] = inbuf[i] ^ state[xorIndex];
    }
    
    rc4->x = x;     
    rc4->y = y;
    
    return;
}
