/*
 * Ether header.
 */
struct ethhdr
{
    u_int8_t	ether_dhost[6];
    u_int8_t	ether_shost[6];
    u_int16_t	ether_type;
};

#ifndef ETHERTYPE_IP
#define ETHERTYPE_IP	0x0800
#endif

/*
 * IP header.
 */
struct iphdr
{
#if	__BYTE_ORDER == __LITTLE_ENDIAN
    u_int8_t	ip_hl	: 4;	/* header length */
    u_int8_t	ip_v	: 4;	/* version */
#elif	__BYTE_ORDER == __BIG_ENDIAN
    u_int8_t	ip_v	: 4;	/* version */
    u_int8_t	ip_hl	: 4;	/* header length */
#else
#error	"Please #define __BYTE_ORDER"
#endif
    u_int8_t	ip_tos;		/* type of service */
    u_int16_t	ip_len;		/* total length */
    u_int16_t	ip_id;		/* identification */
    u_int16_t	ip_off;		/* fragment offset field */
#define	IP_RF	0x8000		/* reserved fragment flag */
#define	IP_DF	0x4000		/* dont fragment flag */
#define	IP_MF	0x2000		/* more fragments flag */
#define	IP_OFFMASK	0x1fff	/* mask for fragmenting bits */
    u_int8_t	ip_ttl;		/* time to live */
    u_int8_t	ip_p;		/* protocol */
    u_int16_t	ip_sum;		/* checksum */
    u_int32_t	ip_src;		/* source address */
    u_int32_t	ip_dst;		/* destination address */
};

#ifndef IPPROTO_TCP
#define IPPROTO_TCP	6
#endif

/*
 * TCP header.
 */
struct tcphdr
{
    u_int16_t	th_sport;	/* source port */
    u_int16_t	th_dport;	/* destination port */
    u_int32_t	th_seq;		/* sequence number */
    u_int32_t	th_ack;		/* acknowledgement number */
#if	__BYTE_ORDER == __LITTLE_ENDIAN
    u_int8_t	th_x2	: 4;	/* (unused) */
    u_int8_t	th_off	: 4;	/* data offset */
#elif	__BYTE_ORDER == __BIG_ENDIAN
    u_int8_t	th_off	: 4;	/* data offset */
    u_int8_t	th_x2	: 4;	/* (unused) */
#else
#error	"Please #define __BYTE_ORDER"    
#endif
    u_int8_t	th_flags;	/* control flags */
#define	TH_FIN	0x01
#define	TH_SYN	0x02
#define	TH_RST	0x04
#define	TH_PUSH	0x08
#define	TH_ACK	0x10
#define	TH_URG	0x20
    u_int16_t	th_win;		/* window */
    u_int16_t	th_sum;		/* checksum */
    u_int16_t	th_urp;		/* urgent pointer */
};

